/**********************************************************************************
 *                                                                                *
 * A sample 'Main' for blocked triangular tridiagonalization (Aasen/Parlett-Reid) *
 *                                                                                *
 * This program computes an NxN random matrix and an N-vector and performs        *
 *  factorization and solve using LAPACK, AASEN or PARLETT-REID.                  *
 * The user can select the partition-size and number of Repetitions               *
 *                                                                                *
 * Usage: ./driver L(apack)/A(asen)/P(arlett-Reid) N parition-size Repetitions    *
 *                                                                                *
 *                                                                                *
 *  Gil Shklarski and Sivan Toledo, May 2007                                      *
 *                                                                                *
 **********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <assert.h>


/* BLAS */
void dcopy_(int * N, double * DX, int * INCX, double * DY,int * INCY);
void dgemv_(char * TRANS,int * M, int * N,double * ALPHA, double * A,int * LDA,
            double * X, int * INCX, double * BETA, double * Y, int * INCY);
double dnrm2_(int * N, double * DX, int * INCX);

/* Random Input Generators */
void createSymmetricMatrix(int N, double **A, long int seedval);
void createRandomVector(int N, double **x, long int seedval);

/* AASEN */
void dstrdf_aasen( int * N, double * A, int *LDA, int * IPIV, 
		   double * buf, double * e1, double * tmp, 
		   double * up_col, double * up_row,
		   int * NB, int * UP_BLOCK_FCT, int * INFO );
void dstrds( int * N, double * A, int *LDA, int * IPIV,int * NB, 
	     double * B, int * LDB, int * NRHS, double * wspace);

/* PARLETT-REID */
void dstrdf_fused( int * N, double * A, int *LDA, int * IPIV, 
		   double * buf, double * e1, double * tmp, int * NB, 
		   int * UP_BLOCK_FCT, int * INFO);
void dstrds_fused( int * N, double * A, int *LDA, int * IPIV,
                   int * NB, double * B, int * LDB, int * NRHS);

/* LAPACK */
int GLOBAL_LAPACK_NB;
int ilaenv_(int * ISPEC, char ** NAME, char ** OPTS, int *N1, int *N2, int *N3, int *N4 ) {
  assert((*ISPEC) == 1);
  
  return GLOBAL_LAPACK_NB;
}

void dsytrf_( char * UPLO, int * N, double * A, int *LDA,
              int * IPIV, double * WORK, int * LWORK, int * INFO );
void dsytrs_( char * UPLO, int * N, int * NRHS, double * A, int * LDA,
              int * IPIV, double * B, int * LDB, int * INFO );


/* Service Routines */

double twonorm(int n, double* v) {
 int one = 1;
  int N = n;
  return dnrm2_(&N,v,&one); 
}

void set_timer() {

  struct itimerval rttimer; 
  struct itimerval old_rttimer;

  rttimer.it_value.tv_sec     = 10000; 
  rttimer.it_value.tv_usec    = 0; 
  rttimer.it_interval.tv_sec  = 0; 
  rttimer.it_interval.tv_usec = 0; 

  setitimer (ITIMER_REAL, &rttimer, &old_rttimer);

}

double _wtime() {
  static struct itimerval tv;

  getitimer(ITIMER_REAL, &tv);

  return (double)tv.it_value.tv_sec + (double)tv.it_value.tv_usec*(double)1.0e-6;
}


int main(int argc, char * argv[]) {
  int N,N2,NB;
  double * A,*x,*b,*copyA, *AX_B, *xhat;
  double wtime;
  int ONE=1;
  double D_M_ONE=-1.0, D_ONE=1.0, D_ZERO = 0.0;
  int * IPIV;
  char charN='N', charL='L';
  int INFO=0;
  int NRHS;
  int LDA;
  double LWORK_DOUBLE=0.0;
  int LWORK=0;
  double * WORK = NULL;
  int RepNUM, Repetitions;
  double * e1, *tmp, *buf, *wspace, *up_row, *up_col;
  int BlockUpdateFactor;
  const char * usageString = "Usage: %s L(apack)/A(asen)/P(arlett-Reid) N parition-size Repetitions\n";
  
  if (argc != 5) {
    fprintf(stderr,usageString,argv[0]);
    exit(-1);
  }
  if ((argv[1][0] != 'L') &&(argv[1][0] != 'A') &&(argv[1][0] != 'P') ) {
    fprintf(stderr, usageString, argv[0]);
    exit(-1);
  }

  set_timer();
  
  N           = atoi(argv[2]);
  NB          = atoi(argv[3]);
  Repetitions = atoi(argv[4]);

  N2 = N*N;
  LDA = N;


  /* Prepare Inputs */
  createSymmetricMatrix(N, &A, 1/*seed*/);
  createRandomVector(N,&x, 2/*seed*/);


  b = (double*)calloc(N,sizeof(double));
  dgemv_(&charN,&N,&N,&D_ONE,A,&LDA,x,&ONE,&D_ZERO,b,&ONE);
  NRHS = 1;

  copyA = (double*)calloc(N*N,sizeof(double));
  dcopy_(&N2,A,&ONE,copyA,&ONE);

  IPIV= (int*)calloc(N,sizeof(int));
  
  if (argv[1][0] == 'A') {

    printf("PARTITIONED_AASEN for N=%d Partition-Size=%d MEM=%d (doubles)",N,NB,4*N+NB*N);

    for (RepNUM=0;RepNUM<Repetitions;RepNUM++) {

      if (RepNUM == 0) {

	e1  =  (double*)calloc(N,sizeof(double));
	tmp =  (double*)calloc(N,sizeof(double));
	buf =  (double*)calloc(N*NB,sizeof(double));
	up_col = (double*)calloc(N,sizeof(double));
	up_row = (double*)calloc(N,sizeof(double));

      }

      if (RepNUM >0) {
	dcopy_(&N2,copyA,&ONE,A,&ONE);
      }
      
    
      /* Performing Factorization */
      wtime = _wtime();
      dstrdf_aasen(&N,A-1,&LDA,IPIV-1,buf-1,e1-1,tmp-1,up_col-1,up_row-1,&NB,&ONE, &INFO );

      printf(" time = %g",wtime-_wtime());
      assert(INFO==0);

      /* Preforming Solve */
      if (RepNUM == 0) { 
	xhat =   (double*)calloc(N,sizeof(double));
	AX_B =   (double*)calloc(N,sizeof(double));
	wspace = (double*)calloc(4*N,sizeof(double));	
      }
      
      dcopy_(&N,b,&ONE,xhat,&ONE);
      dcopy_(&N,b,&ONE,AX_B,&ONE);

      dstrds(&N,A-1,&LDA,IPIV-1,&NB,xhat-1,&N, &NRHS, wspace);
      assert(INFO == 0);

      /* Calculate Residual */
      dgemv_(&charN,&N,&N,&D_M_ONE,copyA,&LDA,xhat,&ONE,&D_ONE,AX_B,&ONE);
      printf(" Residual = %g",dnrm2_(&N,AX_B,&ONE)/dnrm2_(&N,b,&ONE));
    }
    printf("\n");
    
  } else if (argv[1][0] == 'P') {

    printf("PARTITIONED_PARLETT_REID for N=%d Partition-Size=%d MEM=%d (doubles)",N,NB,3*N+NB*N);

    for (RepNUM=0;RepNUM<Repetitions;RepNUM++) {



      if (RepNUM == 0) {

        e1  =  (double*)calloc(N,sizeof(double));
        tmp =  (double*)calloc(N,sizeof(double));
        buf =  (double*)calloc(N*(NB+1),sizeof(double));
      }

      if (RepNUM >0) {
        dcopy_(&N2,copyA,&ONE,A,&ONE);
      }


      /* Performing Factorization */
      wtime = _wtime();
      dstrdf_fused(&N,A-1,&LDA,IPIV-1,buf-1,e1-1,tmp-1,&NB,&ONE, &INFO );
      printf(" time = %g",wtime-_wtime());
      assert(INFO==0);

      /* Preforming Solve */
      if (RepNUM == 0) {
        xhat =   (double*)calloc(N,sizeof(double));
        AX_B =   (double*)calloc(N,sizeof(double));
      }

      dcopy_(&N,b,&ONE,xhat,&ONE);
      dcopy_(&N,b,&ONE,AX_B,&ONE);

      dstrds_fused(&N,A-1,&LDA,IPIV-1,&NB,xhat-1,&N, &NRHS);
      assert(INFO == 0);



      /* Calculate Residual */
      dgemv_(&charN,&N,&N,&D_M_ONE,copyA,&LDA,xhat,&ONE,&D_ONE,AX_B,&ONE);
      printf(" Residual = %2.2e",twonorm(N,AX_B)/twonorm(N,b));
    }
    printf("\n");

  } else {
    
    GLOBAL_LAPACK_NB = NB;



    for (RepNUM=0;RepNUM<Repetitions;RepNUM++) {

      if (RepNUM >0) 
	dcopy_(&N2,copyA,&ONE,A,&ONE);
      

      LWORK = -1;
      dsytrf_(&charL,&N,A,&LDA,IPIV,&LWORK_DOUBLE,&LWORK,&INFO);

      LWORK = (int)LWORK_DOUBLE;
      WORK = (double*)calloc(LWORK,sizeof(double));
      
      if (RepNUM == 0) 
	printf("LAPACK for N=%d Partition-Size=%d MEM=%d (doubles) BLK_UF=1",N,(int)ceil((double)LWORK/(double)N),LWORK);

      /* Performing Factorization */
      wtime = _wtime();
      dsytrf_(&charL,&N,A,&LDA,IPIV,WORK,&LWORK,&INFO);
      printf(" time = %g",wtime-_wtime());
      assert(INFO==0);

      /* Preforming Solve */
      if (RepNUM == 0) { 
	xhat = (double*)calloc(N,sizeof(double));
	AX_B = (double*)calloc(N,sizeof(double));
      }

      dcopy_(&N,b,&ONE,xhat,&ONE);
      dcopy_(&N,b,&ONE,AX_B,&ONE);
      
      dsytrs_(&charL,&N,&NRHS,A,&LDA,IPIV,xhat,&N,&INFO );
      assert(INFO == 0);
            
      
      /* Calculate Residual */
      dgemv_(&charN,&N,&N,&D_M_ONE,copyA,&LDA,xhat,&ONE,&D_ONE,AX_B,&ONE);
      printf(" Residual = %2.2e",twonorm(N,AX_B)/twonorm(N,b));
    }
    printf("\n");


    
  }

  return 0;
}
