
/**********************************************************************************
 *                                                                                *
 * dstrds: Factorization routine for Aasen blocked triangular tridiagonalization  *
 *                                                                                *
 *                                                                                *
 * See 'driver.c' for a propoer usage example.                                    *
 *                                                                                *
 *                                                                                *
 *  Gil Shklarski and Sivan Toledo, May 2007                                      *
 *                                                                                *
 **********************************************************************************/

#include <math.h>
#include <stdio.h>

#define min(x,y) (((x) < (y))? (x) : (y))
#define GIVENS(X,i,j) tA0 = (X)[i]; tA1 = (X)[j]; (X)[i] = c*tA0+s*tA1; (X)[j] =-s*tA0+c*tA1;

/* BLAS Headers */
void dcopy_(int * N, double * DX, int * INCX, double * DY,int * INCY);
void dscal_(int * N, double * DA, double * DX, int *INCX);
void daxpy_(int * N, double * DA, double * DX, int * INCX, double * DY, int * INCY);
void dgemv_(char * TRANS,int * M, int * N,double * ALPHA, double * A,int * LDA,
            double * X, int * INCX, double * BETA, double * Y, int * INCY);
void dgemm_(char * TRANSA, char * TRANSB, int * M,int * N,int *K,
	    double * ALPHA, double * A, int * LDA,
	    double * B, int * LDB, 
	    double * BETA, double * C, int * LDC);
int  idamax_(int * N,  double * DX, int * INCX);
void dswap_(int * N, double * DX, int * INCX, double * DY,int * INCY);
void dger_(int * M, int * N,double * ALPHA,
           double * X,int * INCX, double *Y,int *INCY,
           double *A,int *LDA);




/*
   N    = (INPUT) matrix order.
   NB   = (INPUT) number of columns in a partition
   A    = (INPUT/OUTPUT) matrix stored colomnwise
   LDA  = (INPUT) A's leading dimension
   IPIV = (OUTPUT) pivoting index switch array (of size N)
   wspace  = (INPUT) temporary buffer (of size 4*N)
*/

void dstrds( int * N, double * A, int *LDA, int * IPIV, 
	     int * NB, double * B, int * LDB, int * NRHS,
	     double * wspace) {
  
  int n = *N;
  int nb = *NB;
  int J,i,j,pi;
  int minJplusNBandN;
  double one = 1.0, minus_one = -1.0;
  char charN = 'N',charT = 'T';
  int M;
  int one_int = 1;
  int lda = *LDA;
  int ldb = *LDB;
  double x,y,r,c,s,scale;
  double tA0,tA1;
  int rng1,rng2;
  double * pB;


  if (n == 1) {
    scale = 1/A[1];
    dscal_(NRHS,&scale,B+1,LDB);
    return;
  } else if (n == 2) {
    /* Assume there is only one block */
    if (IPIV[1] != 1)
      dswap_(NRHS,B+1,LDB,B+2,LDB);

    for (j=1;j<=(*NRHS);j++) {
      tA0 = B[1+(j-1)*ldb];
      tA1 = B[2+(j-1)*ldb];

      scale = (A[1]*A[4]-A[3]*A[3]);
      B[1+(j-1)*ldb] = (tA0*A[4]-A[3]*tA1)/scale;
      B[2+(j-1)*ldb] = (A[1]*tA1-A[3]*tA0)/scale;
    }
    if (IPIV[1] != 1)
      dswap_(NRHS,B+1,LDB,B+2,LDB);
    return;
  } 


  /*
    mexPrintf("XXX NRHS = %d, LDB = %d, LDA = %d, NB = %d, nb = %d\n",
    *NRHS, *LDB, *LDA, *NB, nb);
  */
  
  for (J=1; J <= n; J+=nb) {
    minJplusNBandN = min(J+nb,n);
    for (j=0;j< (*NRHS);j++) {
      pB = B+n*j;
      for (i=(J+1); i <= minJplusNBandN; i++) {
	pi = IPIV[i];
	if (i != pi) {
	  tA0 = pB[i];
	  pB[i] = pB[pi];
	  pB[pi] = tA0;
	}
      }
    }

    for (i=(J+1); i <= minJplusNBandN; i++) {
      M = n-i+1;

      /*
      dgemm_(&charN, &charN, &M, NRHS, &one_int,
             &minus_one, A+i+(i-2)*lda, LDA,
             B+i-1, LDB, &one, 
             B+i, LDB);
      */
      dger_(&M,NRHS,&minus_one,
	    A+i+(i-2)*lda, &one_int,
	    B+i-1,LDB,
	    B+i,LDB);

    }
  }
  

  /* Backup data for T */
  for (i=1;i<=n;i++)
    wspace[i] = A[i+(i-1)*lda];
  for (i=1;i<=(n-1);i++)
    wspace[i+n] = A[i+1+(i-1)*lda];
  for (i=1;i<=(n-1);i++)
    wspace[i+2*n-1] = A[i+i*lda];
  for (i=1;i<=(n-2);i++)
    wspace[i+3*n-2] = A[i+(i+1)*lda];

  /* Creating 'T' */
  for (i=1;i<=(n-1);i++)
    A[i+1+(i-1)*lda] = A[i+i*lda];
  for (i=1;i<=(n-2);i++)
    A[i+(i+1)*lda] = 0;

  /* if (n == 1) {
     scale = 1.0/A[1];
     dscal_(NRHS, &scale, B+1, LDB);
     }
     else if (n == 2) {
     x = A[1];
     y = A[2];
     r = sqrt(x*x+y*y);
     c = x/r;
     s = y/r;
     
     for (j=1;j<=(*NRHS);j++)
     GIVENS(B,1+(j-1)*ldb,2+(j-1)*ldb);
     
     } else 
  */
  {

    /*  mexPrintf("XXX Here !\n"); */
    for (i=1;i<=(n-2);i++) {
      x = A[i  +(i-1)*lda];
      y = A[i+1+(i-1)*lda];
      r = sqrt(x*x+y*y);
      c = x/r;
      s = y/r;
      

      /*
      mexPrintf("-- %d BEFORE --\n",i);
      for (I=1;I <= n;I++) {
	for (j=1; j <=n;j++)
	  mexPrintf("%e ",A[I+(j-1)*lda]);
	mexPrintf("\n");
      }
      */

      GIVENS(A,i+(i-1)*lda,i+1+(i-1)*lda);
      GIVENS(A,i+(i  )*lda,i+1+(i  )*lda);
      GIVENS(A,i+(i+1)*lda,i+1+(i+1)*lda);

      /*      mexPrintf("fff11 %d\n",(i+1+(i+1)*lda));*/


      for (j=1;j<=(*NRHS);j++) {
	/* mexPrintf("fff %d\n",i+1+(j-1)*ldb);*/
	GIVENS(B,i+(j-1)*ldb,i+1+(j-1)*ldb);

      }
      /*
      mexPrintf("-- %d AFTER --\n",i);
      for (I=1;I <= n;I++) {
	for (j=1; j <=n;j++)
	  mexPrintf("%e ",A[I+(j-1)*lda]);
	mexPrintf("\n");
      }
      */

    }
    x = A[n-1+(n-2)*lda];
    y = A[n  +(n-2)*lda];
    r = sqrt(x*x+y*y);
    c = x/r;
    s = y/r;

        
    GIVENS(A,n-1+(n-2)*lda,n+(n-2)*lda);
    GIVENS(A,n-1+(n-1)*lda,n+(n-1)*lda);



    for (j=1;j<=(*NRHS);j++) {
      GIVENS(B,n-1+(j-1)*ldb,n+(j-1)*ldb);

    }

  }

 /*
 mexPrintf("After all givens:\n");
  for (i=1;i <= n;i++) {
    for (j=1; j <=n;j++)
      mexPrintf("%e ",A[i+(j-1)*lda]);
    mexPrintf("\n");
  }
 */

  for (i=n;i >= 3; i--) {
    scale = 1.0/A[i+(i-1)*lda];
    dscal_(NRHS, &scale,B+i,LDB);
    for (j=1; j<=(*NRHS);j++) {
      B[i-2+(j-1)*ldb] -= A[i-2+(i-1)*lda]*B[i + (j-1)*ldb];
      B[i-1+(j-1)*ldb] -= A[i-1+(i-1)*lda]*B[i + (j-1)*ldb];
    }
  }

  /*  if (n >= 2) */
  {
    scale = 1.0/A[2+lda];
    dscal_(NRHS,&scale,B+2,LDB);
    scale = -A[1+lda];
    daxpy_(NRHS,&scale,B+2,LDB,B+1,LDB);
  }
  scale = 1.0/A[1];
  dscal_(NRHS,&scale,B+1,LDB);



  /* Restore diags */
  for (i=1;i<=n;i++)
    A[i+(i-1)*lda] = wspace[i];
  for (i=1;i<=(n-1);i++)
    A[i+1+(i-1)*lda] = wspace[i+n];
  for (i=1;i<=(n-1);i++)
    A[i+i*lda] = wspace[i+2*n-1];
  for (i=1;i<=(n-2);i++)
    A[i+(i+1)*lda]= wspace[i+3*n-2];


  /* SOLVE L' XXX */
  J = n;
  if ((n % nb) == 0)
    rng1 = (n/nb-1)*nb+1;
  else
    rng1 = (int)floor((double)n/(double)nb)*nb+1;
  rng2 = n;

  while (rng1 >= 1) {
    /* mexPrintf("rng =[ %d %d] \n",rng1,rng2); */

    for (i=n;i>=(rng1+1);i--) {
      M = min(i-1,rng2)-rng1+1;
      /*      if (M <= 0)
	mexPrintf("Warning\n");
      */
      /* dgemm_(&charT,&charN,&M,NRHS,&one_int,
	        &minus_one, A+i+(rng1-1)*lda, LDA,
		B+i,LDB, &one,
		B+rng1, LDB );
      */

	dger_(&M,NRHS,&minus_one,
	      A+i+(rng1-1)*lda,LDA,
	      B+i,LDB,
	      B+rng1, LDB );
    }
    /*
    x(rng(1):min((i-1),rng(2)),:) = x(rng(1):min((i-1),rng(2)),:)-L(i,rng(1):min(i-1,rng(2)))'*x(i,:);
    */
    
    /*
    i=min(rng2+1,n);
    for (;i>=(rng1+1);i--)
    if (i != IPIV[i])
    dswap_(NRHS,B+i,LDB,B+IPIV[i],LDB);
    */
    
    for (j=0;j<(*NRHS);j++) {
      i=min(rng2+1,n);
      pB = B+n*j;
      for (;i>=(rng1+1);i--) {
	pi = IPIV[i];

	if (i != pi) {
	  tA0 = pB[i];
	  pB[i] = pB[pi];
	  pB[pi] = tA0;
	}
      }
    }


    rng2 = rng1-1;
    rng1 = rng1-nb;

  }

  
}

