function [L,T,P] = partitioned_parlett_reid_fast_code(A,NB)
%
% function [L,T,P] = partitioned_parlett_reid_fast_code(A,NB)
%
% For a symmetric matrix A, finds [L,T,P] s.t. P*A*P' = L*T*L', where L is lower-triangular,
% and T is tridiagonal and symmetric.
%
% NB is the requested number of partitions (defaults to 1).
%
% Performs partitioned 'Parlett-Reid' formulation, with partial pivoting.
%
% This is almost identical to the C-code, in the corresponding C code:
%   * The QR factorization of T is fused with this factorization.
%   * The explicit matrices L, T and P are not computed.
%
%
% Gil Shklarski and Sivan Toledo, 30 May 2007
%


fps = 0;
n = size(A,1);

if (nargin < 2)
    NB = 1;
end

e1 = zeros(n-1,1);

pp = 1:n;

nb = ceil(n/NB);
display(sprintf('Partition columns number = %d',nb));

buf = zeros(n,nb+1);

J=0;
for BL=1:NB
    % factorize block BL

    Jb = J+1;
       
    for J=Jb:min(n,Jb+nb-1)
   
        j = J-Jb+1;
        
        % left looking
        buf(J:n,j) = A(J:n,J)-buf(J:n,1:(j-1))*A(J,Jb:(J-1))';
        fps = fps + (n-J+1)*(2*(j-1));
        
        if (J > Jb)
            tmp = buf(J:n,j) - A(J:n,J-1)*A(J-1,J);
            
            fps = fps + 2*(n-j+1);
        else
            tmp = buf(J:n,j);
        end

        A(J,J) = tmp(1); % because L(j,j)=1
        
        if (J<n)
            tmp(1) = 0;
            tmp(2:end) = tmp(2:end) - e1*A(J,J);
            
            [max_abs_tmp, max_ind] = max(abs(tmp(2:end)));
            max_ind = max_ind + 1;
            max_tmp = tmp(max_ind); 
            
            
            if (abs(max_tmp) ~= max_abs_tmp)
                error('Not possible');
            end

            
            if ((max_ind ~= 2) && (max_abs_tmp ~=0))
                % symmetric permutation
                m1 = 2;
                m2 = max_ind;

                tp=e1(1);
                e1(1) = e1(m2-1);
                e1(m2-1) = tp;

                
                tmp(m2) = tmp(m1);
                tmp(m1) = max_tmp;
             
                m1 = m1 + J-1;
                m2 = m2 + J-1;
                
                buf([m1 m2],1:(m1-Jb)) = buf([m2 m1],1:(m1-Jb));
 
                
                A(m1:(m2-1),m2) = A(m1:(m2-1),m1);
                A(m1:(m2-1),m1) = A(m2,m1:(m2-1))';
                A(m2:n,[m1 m2]) = A(m2:n,[m2 m1]);

                pp(m1) = m2;

                A([m1 m2],Jb:(m1-2))  = A([m2 m1],Jb:(m1-2));

                A([m1 m2],m1) = A([m2 m1],m1);
                A(m2,(m1+1):(m2-1))=A((m1+1):(m2-1),m2)';
                A(m2,m2) = A(m1,m2);
            end
            A((J+1):n,J) = e1;
            
            A(J,J+1) = tmp(2);
            if (tmp(2) ~= 0) 
                e1 = tmp(3:end) / A(J,J+1);
                fps = fps + length(e1);
            else
               e1 = zeros(length(tmp)-1,1);
               e1(J+1) = 1;
            end
            
            
        end
    end

 
    % right looking blocked updated
    if (J < n)
        
         storedA1 = A((J+1):n,J+1);
         A((J+1):n,J+1) = [A(J,J+1); e1* A(J,J+1)];
         
         fps = fps + length(e1);
         
         buf((J+1):n,min(nb,n-Jb+1)+1) = A((J+1):n,J);
         storedA1 = storedA1 - buf((J+1):n,1:(min(nb,n-Jb+1)+1)) * A((J+1),Jb:(J+1))';
    
         fps = fps + 2*min(nb,n-Jb+1)*(n-J);
         
         for J2=(J+1):(nb):n 
             tB = min(nb,n-J2); 
             
             for J3= (J2+1):(J2+tB) 
                 A(J3:(J2+tB),J3) = A(J3:(J2+tB),J3) - ...
                     buf(J3:(J2+tB),1:(min(nb,n-Jb+1)+1)) * A(J3,Jb:(J+1))';
                 fps = fps + 2*(J2+tB-J3+1)*(min(nb,n-Jb+1)+1);
             end             
             A((J2+tB+1):n,(J2+1):(J2+tB)) = A((J2+tB+1):n,(J2+1):(J2+tB)) - ...
                 buf((J2+tB+1):n,1:(min(nb,n-Jb+1)+1)) * A((J2+1):(J2+tB),Jb:(J+1))';
             fps = fps + (n-J2-tB)*2*tB*(min(nb,n-Jb+1)+1);
             
            
         end

         A((J+1):n,J+1) = storedA1;
 
    end
    
end


%%
% The following part is used for the explicit creation of L,T and P and 
% is not performed in the C factorization code

p = 1:n;
for i=1:n
    p([pp(i) i])=p([i pp(i)]);
end


P(1:n,p) = eye(n);
L = tril(A,-1)+eye(length(A));

for J=1:nb:n
    for i=(J+1):min((J+nb),n)
        if (i~=pp(i))
            L([i pp(i)],1:(J-1)) = L([pp(i) i] ,1:(J-1));
        end
    end
end

T = diag(diag(A)) + diag(diag(A,1),1);
T = T + triu(T,1)';

display(sprintf('Flops = %d\n',fps))